{-# LANGUAGE TypeOperators #-}
module Kalah
    ( State(..),
      initState,
      makeMove,
      isTerm,
      isTermPar,
      util,
      utilPar,
      getValidMoves,
      whoseTurn,
      getBoardString,
      whoWon,
      getScore,
      getIsP0Turn
    ) where

import qualified Data.Vector as V
import Control.DeepSeq
import Control.Parallel.Strategies
import Text.Format

data State = State (V.Vector Int) Bool
instance NFData State where
  rnf (State board isP0Turn) = board `deepseq` isP0Turn `seq` ()

-- Initial Kalah state
initState :: State
initState = State (V.generate 14 f) True
  where f i
          | i == 6 || i == 13 = 0
          | otherwise = 4

-- Given a State and a house to pick, outputs the resulting State
makeMove :: State -> Int -> Maybe State
makeMove s@(State board isP0Turn) a = do
  if isValidMove s a
    then do
      let boardFunc i
            | i == a = 0
            | otherwise = board V.! i
      let n = board V.! a
      return $ deposit (a + 1) (-1) boardFunc n isP0Turn
  else do
    Nothing

-- Deposit seeds around the Kalah board
deposit :: Int -> Int -> (Int -> Int) -> Int -> Bool -> State
deposit i prev f n isP0Turn
  | n == 0 = State (V.generate 14 $ steal f prev isP0Turn) nxtIsP0Turn
  | (i == 6 && not isP0Turn) || (i == 13 && isP0Turn) =
      deposit iNxt prev f n isP0Turn
  | otherwise = deposit iNxt i fNew (n - 1) isP0Turn
    where fNew i'
            | i == i' = (f i') + 1
            | otherwise = f i'
          iNxt = mod (i + 1) 14
          nxtIsP0Turn = (prev == 6 && isP0Turn) ||
                        (prev /= 13 && not isP0Turn)

-- Steal seeds from opponent if possible
steal :: (Int -> Int) -> Int -> Bool -> (Int -> Int)
steal f lastPit isP0Turn
  | lastPitOurs && canSteal = fNew 
  | otherwise = f
    where lastPitOurs = (isP0Turn && (0 <= lastPit && lastPit <= 5)) ||
                        (not isP0Turn && (7 <= lastPit && lastPit <= 12))
          seedsInLast = f lastPit
          pitAcross = 12 - lastPit
          seedsAcross = f pitAcross
          canSteal = seedsInLast == 1 && seedsAcross /= 0
          fNew i
            | i == lastPit || i == pitAcross = 0
            | (isP0Turn && i == 6) || (not isP0Turn && i == 13) =
                (f i) + seedsInLast + seedsAcross
            | otherwise = f i

-- Is the game over?
isTerm :: State -> Bool
isTerm (State board _) = (V.sum $ V.slice 0 6 board) == 0 ||
                         (V.sum $ V.slice 7 6 board) == 0

-- Player 0's score - Player 1's score
util :: State -> Int
util (State board _) = (V.sum $ V.slice 0 7 board) -
                       (V.sum $ V.slice 7 7 board)

-- Given a state, what are the valid moves?
getValidMoves :: State -> [Int]
getValidMoves (State board isP0Turn)
  | isP0Turn = filter (houseNotEmpty) [0..5]
  | otherwise = filter (houseNotEmpty) [7..12]
    where houseNotEmpty = houseIsNotEmpty board

-- Given a state and a move, is the move valid?
isValidMove :: State -> Int -> Bool
isValidMove (State board isP0Turn) a
  | isP0Turn = 0 <= a && a <= 5 && houseNotEmpty
  | otherwise = 7 <= a && a <= 12 && houseNotEmpty
    where houseNotEmpty = houseIsNotEmpty board a

-- Is the house empty?
houseIsNotEmpty :: (V.Vector Int) -> Int -> Bool
houseIsNotEmpty board i = board V.! i /= 0

-- Which player's turn is it?
whoseTurn :: State -> Int
whoseTurn (State _ True) = 0
whoseTurn (State _ False) = 1

-- Function to expose isP0Turn inside State
getIsP0Turn :: State -> Bool
getIsP0Turn (State _ isP0Turn) = isP0Turn

-- Pretty prints the board
-- Trust me, it actually looks fine
getBoardString :: State -> String
getBoardString (State board _) = format template $ numsAsStrings
  where template = "        13    12    11    10     9     8\n\
          \ _____ _____ _____ _____ _____ _____ _____ _____\n\
          \|     |     |     |     |     |     |     |     |\n\
          \|     | {12}  | {11}  | {10}  | {9}  | {8}  | {7}  |     |\n\
          \| {13}  |_____|_____|_____|_____|_____|_____| {6}  |\n\
          \|     |     |     |     |     |     |     |     |\n\
          \|     | {0}  | {1}  | {2}  | {3}  | {4}  | {5}  |     |\n\
          \|_____|_____|_____|_____|_____|_____|_____|_____|\n\
          \         1     2     3     4     5     6"
        numsAsStrings = fmap (padNum . show) $ V.toList board
        padNum i
          | length i == 2 = i
          | otherwise = " " ++ i

-- Who won?
whoWon :: State -> String
whoWon state
  | utilVal > 0 = "Player 0 wins!"
  | utilVal < 0 = "Player 1 wins!"
  | otherwise = "Tie!"
    where utilVal = util state

-- Prints each player's score
getScore :: State -> String
getScore (State board _) =
  "Player 0 score: " ++ p0Score ++ "  -  Player 1 score: " ++ p1Score
  where p0Score = show $ V.sum $ V.slice 0 7 board
        p1Score = show $ V.sum $ V.slice 7 7 board

-- isTerm and util, but in parallel
isTermPar :: State -> Bool
isTermPar (State board _) = runEval $ do
  a <- rpar $ (V.sum $ V.slice 0 6 board) == 0
  b <- rpar $ (V.sum $ V.slice 7 6 board) == 0
  _ <- rseq $ a || b
  return $ a || b

utilPar :: State -> Int
utilPar (State board _) = runEval $ do
  a <- rpar $ V.sum $ V.slice 0 7 board
  b <- rpar $ V.sum $ V.slice 7 7 board
  _ <- rseq a
  _ <- rseq b
  return $ a - b
