module MinMax
    ( computeAIMove,
      MinMaxConfig(..),
      MinMaxMode(..)
    ) where

import Kalah
import Data.List(maximumBy, minimumBy)
import Control.Parallel.Strategies

data MinMaxConfig = MinMaxConfig { mmMode :: MinMaxMode
                                 , depthLimit :: Int
                                 , parLimit :: Int }
data MinMaxMode = SeqMM | SeqAB | ParThenMM | ParThenAB | YoungBros
  deriving (Enum)

posInfty :: Int
posInfty = 99999999

negInfty :: Int
negInfty = -posInfty

genSuccs :: State -> [(Int, State)]
genSuccs s = [(a, unwrapState $ makeMove s a) | a <- getValidMoves s]
  where unwrapState (Just s') = s'
        unwrapState Nothing = error "Something went wrong during MinMax"

-- Sequential MM
runMM :: Int -> State -> (Int, Int)
runMM k s
  | k == 0 || isTerm s = (util s, -1)
  | getIsP0Turn s = maximumBy valMoveOrdering childrenMMRes
  | otherwise = minimumBy valMoveOrdering childrenMMRes
    where succs = genSuccs s
          childrenMMRes = [(fst $ runMM (k - 1) s', a) | (a, s') <- succs]

valMoveOrdering :: (Int, Int) -> (Int, Int) -> Ordering
valMoveOrdering (v0,_) (v1,_) = compare v0 v1

-- Sequential AB
runAB :: Int -> State -> (Int, Int) -> (Int, Int)
runAB k s (a, b)
  | k == 0 || isTerm s = (util s, -1)
  | getIsP0Turn s = maxAB (k - 1) (a, b) (negInfty, -1) succs
  | otherwise = minAB (k - 1) (a, b) (posInfty, -1) succs
    where succs = genSuccs s

maxAB :: Int -> (Int, Int) -> (Int, Int) -> [(Int, State)] -> (Int, Int)
maxAB _ _ (v, move) [] = (v, move)
maxAB k (a, b) (v, move) ((sMove, s):succs)
  | v' > b = (v', move')
  | otherwise = maxAB k (a', b) (v', move') succs
    where (v2, _) = runAB k s (a, b)
          (v', move', a')
            | v2 > v = (v2, sMove, max a v)
            | otherwise = (v, move, a)

minAB :: Int -> (Int, Int) -> (Int, Int) -> [(Int, State)] -> (Int, Int)
minAB _ _ (v, move) [] = (v, move)
minAB k (a, b) (v, move) ((sMove, s):succs)
  | v' <= a = (v', move')
  | otherwise = minAB k (a, b') (v', move') succs
    where (v2, _) = runAB k s (a, b)
          (v', move', b')
            | v2 < v = (v2, sMove, min b v)
            | otherwise = (v, move, b)

-- Run par MM for parDepth layers, then seq MM for deeper layers
runParThenMM :: Int -> Int -> State -> (Int, Int)
runParThenMM parDepth k s
  | k == 0 || isTerm s = (util s, -1)
  | parDepth == 0 = runMM k s
  | getIsP0Turn s = maximumBy valMoveOrdering childrenMMRes
  | otherwise = minimumBy valMoveOrdering childrenMMRes
    where succs = genSuccs s
          runMMOnSucc (a, s') =
            (fst $ runParThenMM (parDepth - 1) (k - 1) s', a)
          childrenMMRes = parMap rdeepseq runMMOnSucc succs

-- Run par MM for parDepth layers, then seq AB for deeper layers
runParThenAB :: Int -> Int -> State -> (Int, Int)
runParThenAB parDepth k s
  | k == 0 || isTerm s = (util s, -1)
  | parDepth == 0 = runAB k s (negInfty, posInfty)
  | getIsP0Turn s = maximumBy valMoveOrdering childrenMMRes
  | otherwise = minimumBy valMoveOrdering childrenMMRes
    where succs = genSuccs s
          runMMOnSucc (a, s') =
            (fst $ runParThenAB (parDepth - 1) (k - 1) s', a)
          childrenMMRes = parMap rdeepseq runMMOnSucc succs

-- Run seq AB on first child, prune, then recurse on rest of children
-- Once parDepth is reached, only seq AB is used
runYoungBros :: Int -> Int -> State -> (Int, Int) -> (Int, Int)
runYoungBros parDepth k s (a, b)
  | k == 0 || isTerm s = (util s, -1)
  | parDepth == 0 = runAB k s (a, b)
  | getIsP0Turn s = runYoungBrosMax (parDepth - 1) (k - 1) (a, b) succs
  | otherwise = runYoungBrosMin (parDepth - 1) (k - 1) (a, b) succs
    where succs = genSuccs s

runYoungBrosMax :: Int -> Int -> (Int, Int) -> [(Int, State)] -> (Int, Int)
runYoungBrosMax _ _ _ [] = error "Something went wrong during MinMax"
runYoungBrosMax parDepth k (a, b) ((house, s):succs)
  | v > b = (v, move)
  | otherwise = maximumBy valMoveOrdering ((v, house):childrenMMRes)
    where (v, move) = runAB k s (a, b)
          a' = max a v
          runYoungBrosOnSucc (house', s') =
            (fst $ runYoungBros parDepth k s' (a', b), house')
          childrenMMRes = parMap rdeepseq runYoungBrosOnSucc succs

runYoungBrosMin :: Int -> Int -> (Int, Int) -> [(Int, State)] -> (Int, Int)
runYoungBrosMin _ _ _ [] = error "Something went wrong during MinMax"
runYoungBrosMin parDepth k (a, b) ((house, s):succs)
  | v <= a = (v, move)
  | otherwise = minimumBy valMoveOrdering ((v, house):childrenMMRes)
    where (v, move) = runAB k s (a, b)
          b' = min b v
          runYoungBrosOnSucc (house', s') =
            (fst $ runYoungBros parDepth k s' (a, b'), house')
          childrenMMRes = parMap rdeepseq runYoungBrosOnSucc succs

computeAIMove :: MinMaxConfig -> State -> Int
computeAIMove (MinMaxConfig mode' k parDepth) s
  | k < 0 = error "Invalid search depth limit."
  | otherwise =
    case mode' of
      SeqMM -> snd $ runMM k s
      SeqAB -> snd $ runAB k s (negInfty, posInfty)
      ParThenMM -> snd $ runParThenMM parDepth k s
      ParThenAB -> snd $ runParThenAB parDepth k s
      YoungBros -> snd $ runYoungBros parDepth k s (negInfty, posInfty)
